const int testDataAligned = TEST_DATA_ALIGNED;

#if (TEST_DATA_ALIGNED == 1)
#   pragma align 4
#endif

#if (TEST_PHASE == 1)
#   pragma section farrom "memcpy_uncached_pflash0"
#   pragma section farnoclear "memcpy_bss_uncached_lmuram"
#elif (TEST_PHASE == 2)
#   pragma section farrom "memcpy_uncached_pflash0"
#   pragma section farnoclear "memcpy_bss_uncached_lmuram"
#elif (TEST_PHASE == 3)
#   pragma section farrom "memcpy_cached_pflash0"
#   pragma section farnoclear "memcpy_bss_uncached_lmuram"
#elif (TEST_PHASE == 4)
#   pragma section farrom "memcpy_cached_pflash0"
#   pragma section farnoclear "memcpy_bss_lmuram"
#elif (TEST_PHASE >= 5)
#   pragma section farrom "memcpy_cached_pflash0"
#   pragma section farnoclear "memcpy_bss_dspr0"
#else
#   error "No test phase specified (macro TEST_PHASE undefined)"
#endif

char const pSrc[MAX_NUM_BYTES] = "mySourceContent!";
char pDest[MAX_NUM_BYTES];
unsigned short nBytes;

#pragma section farrom restore
#pragma section farnoclear restore
#if (TEST_DATA_ALIGNED == 1)
#   pragma align restore
#endif

const int testPhase = TEST_PHASE; /* allows easy checking which TEST_PHASE an ELF was compiled for when debugging */

#if (TEST_PHASE < 6)
#   pragma optimize O0 /* Disable TASKING compiler optimizations to start with worst performing implementation of memcpy */
#else
#   pragma optimize O3
#endif

#if (TEST_PHASE == 1 || TEST_PHASE == 9)
#   pragma section code "memcpy_uncached_pflash0"
#elif (TEST_PHASE == 2)
#   pragma section code "memcpy_cached_pflash0"
#elif (TEST_PHASE == 3)
#   pragma section code "memcpy_cached_pflash0"
#elif (TEST_PHASE == 4)
#   pragma section code "memcpy_cached_pflash0"
#elif (TEST_PHASE >= 5)
#   pragma section code "memcpy_cached_pflash0"
#else
#   error "No test phase specified (macro TEST_PHASE undefined)"
#endif

#if (TEST_PHASE < 7)
void* memcpy(void* pDest, void const* pSrc, unsigned short n)
{
    /* char pointers allows copying with byte granularity */
    char* pD = pDest;
    char const* pS = pSrc;

    while(n--)
    {
        *pD++ = *pS++;
    }
    return pDest;
}
#else /* if (TEST_PHASE < 7) */
void* memcpy(void* pDest, void const* pSrc, unsigned short n)
{
    unsigned int wordCount = n >> 2u;

    /* Check pDest and pSrc for word alignment. at the same
       time check whether n is a multiple of 4 */
    if(0u==(((unsigned int)pDest|(unsigned int)pSrc|n)&3u))
    {
        /* Now use word (1 word = 4 bytes) pointers */
        unsigned int* pD = (unsigned int*)pDest;
        unsigned int const* pS = (unsigned int const*)pSrc;

        while(0u != wordCount--)
        {
            *pD++ = *pS++; /* Copy 4 bytes at a time */
        }
    }
    else
    {
        char* pD = pDest;
        char const* pS = pSrc;
#   if (TEST_PHASE == 7)
        /* use old implementation of memcpy */

        while(n--)
        {
            *pD++ = *pS++;
        }
#   else
        /* use Duff's device for non-aligned stuff */

        /* Calculate how many number of iterations are left which are a multiple of four.
           This is after the first entrance to the switch-case which performs the byte copying
           which are not a multiple of 4..i.e the trailing bytes.
           RSH 2 = Divide by 4. Essentially we are loop-unrolling at a depth of 4 here.. */
        unsigned int byteAllignedCount = ( n + 3u ) >> 2u;

#       pragma warning 553 /* surpress waringn "no 'break' or comment before case label" */
        switch ( n % 4u )
        {
        case 0u:
            do
            {
                *pD++ = *pS++;
                case 3u: *pD++ = *pS++;
                case 2u: *pD++ = *pS++;
                case 1u: *pD++ = *pS++;
            }
            /* Pre-decrement byteAllignedCount. This while loop condition is always tested  */
            while ( 0u < --byteAllignedCount );
        }
#       pragma warning restore
#   endif
    }
    return pDest;
}
#endif /* else (TEST_PHASE < 7) */

#pragma section code restore
#pragma endoptimize

/*-----------------------------------------------------------------------------------*/
